<Query Kind="Program">
  <Namespace>System.Threading.Tasks</Namespace>
</Query>

async static Task Main()
{
	(await GetCachedContentAsync (new Uri ("http://foo"))).Dump();
}

static Task<string> GetContent (Uri uri) => Task.Delay (1000).ContinueWith (_ => "slow");

static readonly Func<Uri, Task<string>> GetCachedContentAsync = Memoize<Uri, Task<string>> (GetContent);

static Func<TKey,TValue> Memoize<TKey,TValue> (Func<TKey,TValue> getValueFunc)
{
	var cache = new Dictionary<TKey,TValue>();
	return (key =>
	{
		TValue result;
		lock (cache)
		{
			if (cache.TryGetValue (key, out result)) return result;
			return cache [key] = getValueFunc (key);
		}
	});
}