<Query Kind="Program">
  <Namespace>System.Threading.Tasks</Namespace>
</Query>

async static Task Main()
{
	(await GetCachedContentAsync (new Uri ("http://foo"))).Dump();
}

static Task<string> GetContent (Uri uri) => Task.Delay (1000).ContinueWith (_ => "slow");

static Memoizer<Uri, Task<string>> _cache = new Memoizer<Uri, Task<string>> (GetContent);

static Task<string> GetCachedContentAsync (Uri uri) => _cache.GetValue (uri);

class Memoizer<TKey,TValue>
{
	readonly Dictionary<TKey,TValue> _cache = new Dictionary<TKey,TValue>();
	readonly Func<TKey,TValue> _getValueFunc;

	public Memoizer (Func<TKey,TValue> getValueFunc) { _getValueFunc = getValueFunc; }

	public TValue GetValue (TKey key)
	{
		TValue result;
		lock (_cache)
		{
			if (_cache.TryGetValue (key, out result)) return result;
			return _cache [key] = _getValueFunc (key);
		}
	}
}