<Query Kind="Program">
  <Namespace>System.Threading.Tasks</Namespace>
</Query>

async static Task Main()
{
	(await GetCachedContentAsync (new Uri ("http://foo"))).Dump();	
}

static Task<string> GetContent (Uri uri) => Task.Delay (1000).ContinueWith (_ => "slow");

static Memoizer<Uri, Task<string>> _cache = new GetPageContentMemoizer();

static Task<string> GetCachedContentAsync (Uri uri) => _cache.GetValue (uri);

abstract class Memoizer<TKey,TValue>
{
	readonly Dictionary<TKey,TValue> _cache = new Dictionary<TKey,TValue>();

	protected abstract TValue GetValueCore (TKey key);

	public TValue GetValue (TKey key)
	{
		TValue result;
		lock (_cache)
		{
			if (_cache.TryGetValue (key, out result)) return result;
			return _cache [key] = GetValueCore (key);
		}
	}
}

class GetPageContentMemoizer : Memoizer<Uri, Task<string>>
{
	protected override Task<string> GetValueCore (Uri key) => GetContent (key);
}