<Query Kind="Program">
  <Namespace>System.Threading.Tasks</Namespace>
  <Namespace>System.IO.MemoryMappedFiles</Namespace>
</Query>

class Foo : AsyncDisposable
{
	readonly MemoryMappedFile _mmFile;
	object _somethingImportant;	

	public async Task Method1()
	{
		using (DeferDisposal())
		{
			#region implementation
			#endregion
		}
	}

	public async Task Method2()
	{
		using (DeferDisposal())
		{
			#region implementation
			#endregion
		}
	}

	public async Task Method3()
	{
		using (DeferDisposal())
		{
			#region implementation
			#endregion
		}
	}
	
	public override async Task DisposeAsync()
	{
		// Await for all outstanding operations to complete.
		await base.DisposeAsync().ConfigureAwait (false);
		
		_mmFile?.Dispose();
		_somethingImportant = null;
	}
}

#region AsyncDisposable

public class AsyncDisposable : IDisposable
{
	readonly object _locker = new object();
	readonly CountdownEvent _deferralEvent = new CountdownEvent (1);
	bool _disposeRequested;

	/// <summary>Suspends disposal until the token that it returns is disposed.
	/// If already disposed, throws an ObjectDisposedException.</summary>
	public IDisposable DeferDisposal()
	{
		if (!_deferralEvent.TryAddCount ()) throw new ObjectDisposedException (GetType ().Name);
		return Disposable.Create (() => _deferralEvent.Signal());
	}

	/// <summary>Starts disposal. If disposal has been deferred, disposal will not start 
	/// until the deferral tokens have been released.</summary>
	public virtual async Task DisposeAsync()
	{
		lock (_locker)
		{
			if (_disposeRequested) return;
			_disposeRequested = true;
		}

		_deferralEvent.Signal();
		await _deferralEvent.WaitHandle.ToTask().ConfigureAwait (false);
	}

	public async void Dispose() => await DisposeAsync();
}

#endregion
#region Helper Types

class Disposable : IDisposable
{
	public static IDisposable Create (Action onDispose) => new Disposable (onDispose);

	Action _onDispose;

	Disposable (Action onDispose) { _onDispose = onDispose; }

	public void Dispose()
	{
		Action todo;
		lock (this)
		{
			todo = _onDispose;
			_onDispose = null;
		}
		if (todo != null) todo();
	}
}

static class Extensions
{
	public static Task ToTask (this WaitHandle waitHandle,
		CancellationToken cancelToken = default(CancellationToken), int timeout = -1)
	{
		var tcs = new TaskCompletionSource<object> ();

		RegisteredWaitHandle token = null;

		// There's a bug in RegisterWaitForSingleObject where calling token.Unregister on an AutoResetEvent can cause
		// subsequent invocations to trigger immediately. We'll leave it instead to the GC to do the job.

		var cancelTokenReg = cancelToken.Register (() => tcs.TrySetCanceled ());

		if (cancelToken.IsCancellationRequested) return tcs.Task;

		token = ThreadPool.RegisterWaitForSingleObject (
			waitHandle,
			(state, timedOut) =>
			{
				cancelTokenReg.Dispose ();
				if (timedOut) tcs.TrySetException (new TimeoutException ());
				else tcs.TrySetResult (null);
				GC.KeepAlive (token);
			},
			null,
			timeout,
			true);

		return tcs.Task;
	}
}

#endregion