<Query Kind="Program">
  <Namespace>System.Threading.Tasks</Namespace>
  <Namespace>System.Net</Namespace>
</Query>

public static class Extensions
{
	// http://blogs.msdn.com/b/pfxteam/archive/2012/08/15/implementing-then-with-await.aspx

	public static async Task<TNewResult> Then<TResult, TNewResult> (
		this Task<TResult> task, Func<TResult, Task<TNewResult>> continuation)
			=> await continuation (await task);

#region Optimized, plus support for nongeneric tasks

	public static async Task<TNewResult> Then_<TResult, TNewResult>(this Task<TResult> task, Func<TResult, Task<TNewResult>> continuation)
		=> await continuation (await task.ConfigureAwait (false)).ConfigureAwait (false);

	public static async Task Then<TResult> (this Task<TResult> task, Func<TResult, Task> continuation)
		=> await continuation (await task);

	public static async Task<TNewResult> Then<TNewResult> (this Task task, Func<Task<TNewResult>> continuation)
	{
		await task.ConfigureAwait (false);
		return await continuation().ConfigureAwait (false);
	}

	public static async Task Then (this Task task, Func<Task> continuation)
	{
		await task.ConfigureAwait (false);
		await continuation().ConfigureAwait (false);
	}
	
#endregion	
}