<Query Kind="Program">
  <Namespace>System.Threading.Tasks</Namespace>
</Query>

void Main()
{
	
}

class AsyncLock
{
	Task _tail = Task.FromResult (true);
	
	public Task<TResult> Enter<TResult> (Func<Task<TResult>> func)
	{
		var result = _tail.Then (func);
		_tail = result;
		return result;
	}
}

#region Extensions
public static class Extensions
{
	public static Task<T> Catch<T, TError>(this Task<T> task, Func<TError, T> onError) where TError : Exception
	{
		var tcs = new TaskCompletionSource<T> ();

		task.ContinueWith (ant =>
		{
			if (task.IsFaulted && task.Exception.InnerException is TError)
				tcs.SetResult (onError ((TError)task.Exception.InnerException));
			else if (ant.IsCanceled)
				tcs.SetCanceled ();
			else if (task.IsFaulted)
				tcs.SetException (ant.Exception.InnerException);
			else
				tcs.SetResult (ant.Result);
		});
		return tcs.Task;
	}

	public static Task<T> Catch<T, TError>(this Task<T> task, T onError) where TError : Exception
	{
		return task.Catch<T, TError> (ex => onError);
	}

	public static Task<T> Catch<T, TError>(this Task task, T onError) where TError : Exception
	{
		return task.ToTaskOfT<T>().Catch<T, TError> (onError);
	}

	public static Task Catch (this Task task)
	{
		return task.Catch<object, Exception> (null);
	}

	public static Task<T> Catch<T>(this Task<T> task, T valueIfError = default(T))
	{
		return task.Catch<T, Exception> (valueIfError);
	}

	public async static Task<T> ToTaskOfT<T>(this Task t)
	{
		await t;
		return default(T);
	}

	// Stephen Toub's monadic bind
	public static async Task Then (this Task task, Func<Task> continuation)
	{
		await task;
		await continuation();
	}

	public static async Task<TNewResult> Then<TNewResult>(this Task task, Func<Task<TNewResult>> continuation)
	{
		await task;
		return await continuation();
	}

	public static async Task Then<TResult>(this Task<TResult> task, Func<TResult, Task> continuation)
	{
		await continuation (await task);
	}

	public static async Task<TNewResult> Then<TResult, TNewResult>(this Task<TResult> task, Func<TResult, Task<TNewResult>> continuation)
	{
		return await continuation (await task);
	}
}
#endregion